@@ -1,7 +1,9 @@
 #!/usr/bin/python
 # Impacket - Collection of Python classes for working with network protocols.
 #
-# SECUREAUTH LABS. Copyright (C) 2021 SecureAuth Corporation. All rights reserved.
+# Copyright Fortra, LLC and its affiliated companies 
+#
+# All rights reserved.
 #
 # This software is provided under a slightly modified version
 # of the Apache Software License. See the accompanying LICENSE file
@@ -17,103 +19,17 @@
 #   Structure
 #
 
-from __future__ import division
-from __future__ import print_function
 import argparse
 import sys
-import os
 import logging
 
 from impacket.examples import logger
+from impacket.examples.mssqlshell import SQLSHELL
 from impacket.examples.utils import parse_target
 from impacket import version, tds
 
+
 if __name__ == '__main__':
-    import cmd
-
-    class SQLSHELL(cmd.Cmd):
-        def __init__(self, SQL):
-            cmd.Cmd.__init__(self)
-            self.sql = SQL
-            self.prompt = 'SQL> '
-            self.intro = '[!] Press help for extra shell commands'
-
-        def do_help(self, line):
-            print("""
-     lcd {path}                 - changes the current local directory to {path}
-     exit                       - terminates the server process (and this session)
-     enable_xp_cmdshell         - you know what it means
-     disable_xp_cmdshell        - you know what it means
-     xp_cmdshell {cmd}          - executes cmd using xp_cmdshell
-     sp_start_job {cmd}         - executes cmd using the sql server agent (blind)
-     ! {cmd}                    - executes a local shell cmd
-     """) 
-
-        def do_shell(self, s):
-            os.system(s)
-
-        def do_xp_cmdshell(self, s):
-            try:
-                self.sql.sql_query("exec master..xp_cmdshell '%s'" % s)
-                self.sql.printReplies()
-                self.sql.colMeta[0]['TypeData'] = 80*2
-                self.sql.printRows()
-            except:
-                pass
-
-        def sp_start_job(self, s):
-            try:
-                self.sql.sql_query("DECLARE @job NVARCHAR(100);"
-                                   "SET @job='IdxDefrag'+CONVERT(NVARCHAR(36),NEWID());"
-                                   "EXEC msdb..sp_add_job @job_name=@job,@description='INDEXDEFRAG',"
-                                   "@owner_login_name='sa',@delete_level=3;"
-                                   "EXEC msdb..sp_add_jobstep @job_name=@job,@step_id=1,@step_name='Defragmentation',"
-                                   "@subsystem='CMDEXEC',@command='%s',@on_success_action=1;"
-                                   "EXEC msdb..sp_add_jobserver @job_name=@job;"
-                                   "EXEC msdb..sp_start_job @job_name=@job;" % s)
-                self.sql.printReplies()
-                self.sql.printRows()
-            except:
-                pass
-
-        def do_lcd(self, s):
-            if s == '':
-                print(os.getcwd())
-            else:
-                os.chdir(s)
-    
-        def do_enable_xp_cmdshell(self, line):
-            try:
-                self.sql.sql_query("exec master.dbo.sp_configure 'show advanced options',1;RECONFIGURE;"
-                                   "exec master.dbo.sp_configure 'xp_cmdshell', 1;RECONFIGURE;")
-                self.sql.printReplies()
-                self.sql.printRows()
-            except:
-                pass
-
-        def do_disable_xp_cmdshell(self, line):
-            try:
-                self.sql.sql_query("exec sp_configure 'xp_cmdshell', 0 ;RECONFIGURE;exec sp_configure "
-                                   "'show advanced options', 0 ;RECONFIGURE;")
-                self.sql.printReplies()
-                self.sql.printRows()
-            except:
-                pass
-
-        def default(self, line):
-            try:
-                self.sql.sql_query(line)
-                self.sql.printReplies()
-                self.sql.printRows()
-            except:
-                pass
-         
-        def emptyline(self):
-            pass
-
-        def do_exit(self, line):
-            return True
-
     # Init the example's logger theme
     logger.init()
     print(version.BANNER)
@@ -121,11 +37,11 @@ if __name__ == '__main__':
     parser = argparse.ArgumentParser(add_help = True, description = "TDS client implementation (SSL supported).")
 
     parser.add_argument('target', action='store', help='[[domain/]username[:password]@]<targetName or address>')
-    parser.add_argument('-port', action='store', default='1433', help='target MSSQL port (default 1433)')
     parser.add_argument('-db', action='store', help='MSSQL database instance (default None)')
     parser.add_argument('-windows-auth', action='store_true', default=False, help='whether or not to use Windows '
                                                                                   'Authentication (default False)')
     parser.add_argument('-debug', action='store_true', help='Turn DEBUG output ON')
+    parser.add_argument('-show', action='store_true', help='show the queries')
     parser.add_argument('-file', type=argparse.FileType('r'), help='input file with commands to execute in the SQL shell')
 
     group = parser.add_argument_group('authentication')
@@ -137,13 +53,21 @@ if __name__ == '__main__':
                        'ones specified in the command line')
     group.add_argument('-aesKey', action="store", metavar = "hex key", help='AES key to use for Kerberos Authentication '
                                                                             '(128 or 256 bits)')
+
+    group = parser.add_argument_group('connection')
+
     group.add_argument('-dc-ip', action='store',metavar = "ip address",  help='IP Address of the domain controller. If '
                        'ommited it use the domain part (FQDN) specified in the target parameter')
+    group.add_argument('-target-ip', action='store', metavar = "ip address",
+                       help='IP Address of the target machine. If omitted it will use whatever was specified as target. '
+                            'This is useful when target is the NetBIOS name and you cannot resolve it')
+    group.add_argument('-port', action='store', default='1433', help='target MSSQL port (default 1433)')
+
 
     if len(sys.argv)==1:
         parser.print_help()
         sys.exit(1)
- 
+
     options = parser.parse_args()
 
     if options.debug is True:
@@ -153,7 +77,7 @@ if __name__ == '__main__':
     else:
         logging.getLogger().setLevel(logging.INFO)
 
-    domain, username, password, address = parse_target(options.target)
+    domain, username, password, remoteName = parse_target(options.target)
 
     if domain is None:
         domain = ''
@@ -162,10 +86,13 @@ if __name__ == '__main__':
         from getpass import getpass
         password = getpass("Password:")
 
+    if options.target_ip is None:
+        options.target_ip = remoteName
+
     if options.aesKey is not None:
         options.k = True
 
-    ms_sql = tds.MSSQL(address, int(options.port))
+    ms_sql = tds.MSSQL(options.target_ip, int(options.port), remoteName)
     ms_sql.connect()
     try:
         if options.k is True:
@@ -179,7 +106,7 @@ if __name__ == '__main__':
         logging.error(str(e))
         res = False
     if res is True:
-        shell = SQLSHELL(ms_sql)
+        shell = SQLSHELL(ms_sql, options.show)
         if options.file is None:
             shell.cmdloop()
         else:
