@@ -1,7 +1,9 @@
 #!/usr/bin/python
 # Impacket - Collection of Python classes for working with network protocols.
 #
-# SECUREAUTH LABS. Copyright (C) 2020 SecureAuth Corporation. All rights reserved.
+# Copyright Fortra, LLC and its affiliated companies 
+#
+# All rights reserved.
 #
 # This software is provided under a slightly modified version
 # of the Apache Software License. See the accompanying LICENSE file
@@ -43,6 +45,7 @@ if __name__ == '__main__':
     parser.add_argument('-ip', '--interface-address', action='store', default='0.0.0.0', help='ip address of listening interface')
     parser.add_argument('-port', action='store', default='445', help='TCP port for listening incoming connections (default 445)')
     parser.add_argument('-smb2support', action='store_true', default=False, help='SMB2 Support (experimental!)')
+    parser.add_argument('-outputfile', action='store', default=None, help='Output file to log smbserver output messages')
 
     if len(sys.argv)==1:
         parser.print_help()
@@ -70,6 +73,10 @@ if __name__ == '__main__':
 
     server = smbserver.SimpleSMBServer(listenAddress=options.interface_address, listenPort=int(options.port))
 
+    if options.outputfile:
+        logging.info('Switching output to file %s' % options.outputfile)
+        server.setLogFile(options.outputfile)
+
     server.addShare(options.shareName.upper(), options.sharePath, comment)
     server.setSMB2Support(options.smb2support)
 
@@ -97,9 +104,5 @@ if __name__ == '__main__':
     # e.g. server.setSMBChallenge('12345678abcdef00')
     server.setSMBChallenge('')
 
-    # If you don't want log to stdout, comment the following line
-    # If you want log dumped to a file, enter the filename
-    server.setLogFile('')
-
     # Rock and roll
     server.start()
