@@ -1,21 +1,29 @@
-# Copyright (c) 2003-2016 CORE Security Technologies
+# Impacket - Collection of Python classes for working with network protocols.
 #
-# This software is provided under under a slightly modified version
+# Copyright Fortra, LLC and its affiliated companies 
+#
+# All rights reserved.
+#
+# This software is provided under a slightly modified version
 # of the Apache Software License. See the accompanying LICENSE file
 # for more information.
 #
-# Author: Alberto Solino (@agsolino)
-#
 # Description:
 #   Kerberos Credential Cache format implementation
 #   based on file format described at:
-#   http://repo.or.cz/w/krb5dissect.git/blob_plain/HEAD:/ccache.txt
+#   https://repo.or.cz/w/krb5dissect.git/blob_plain/HEAD:/ccache.txt
 #   Pretty lame and quick implementation, not a fun thing to do
 #   Contribution is welcome to make it the right way
 #
-
+# Author:
+#   Alberto Solino (@agsolino)
+#
+from __future__ import division
+from __future__ import print_function
 from datetime import datetime
+import os
 from struct import pack, unpack, calcsize
+from six import b, PY2
 
 from pyasn1.codec.der import decoder, encoder
 from pyasn1.type.univ import noValue
@@ -23,9 +31,16 @@ from binascii import hexlify
 
 from impacket.structure import Structure
 from impacket.krb5 import crypto, constants, types
-from impacket.krb5.asn1 import AS_REP, seq_set, TGS_REP, EncTGSRepPart, EncASRepPart, Ticket
+from impacket.krb5.asn1 import AS_REP, seq_set, TGS_REP, EncTGSRepPart, EncASRepPart, Ticket, KRB_CRED, \
+    EncKrbCredPart, KrbCredInfo, seq_set_iter
+from impacket.krb5.types import KerberosTime
 from impacket import LOG
 
+try:
+    FileNotFoundError
+except NameError:
+    FileNotFoundError = IOError
+
 DELTA_TIME = 1
 
 class Header(Structure):
@@ -52,7 +67,22 @@ class CountedOctetString(Structure):
     def prettyPrint(self, indent=''):
         return "%s%s" % (indent, hexlify(self['data']))
 
-class KeyBlock(Structure):
+
+class KeyBlockV3(Structure):
+    structure = (
+        ('keytype','!H=0'),
+        ('etype','!H=0'),
+        ('etype2', '!H=0'),  # Version 3 repeats the etype
+        ('keylen','!H=0'),
+        ('_keyvalue','_-keyvalue','self["keylen"]'),
+        ('keyvalue',':'),
+    )
+
+    def prettyPrint(self):
+        return "Key: (0x%x)%s" % (self['keytype'], hexlify(self['keyvalue']))
+
+
+class KeyBlockV4(Structure):
     structure = (
         ('keytype','!H=0'),
         ('etype','!H=0'),
@@ -72,10 +102,10 @@ class Times(Structure):
         ('renew_till','!L=0'),
     )
     def prettyPrint(self, indent = ''):
-        print "%sAuth : %s" % (indent, datetime.fromtimestamp(self['authtime']).isoformat())
-        print "%sStart: %s" % (indent, datetime.fromtimestamp(self['starttime']).isoformat())
-        print "%sEnd  : %s" % (indent, datetime.fromtimestamp(self['endtime']).isoformat())
-        print "%sRenew: %s" % (indent, datetime.fromtimestamp(self['renew_till']).isoformat())
+        print(("%sAuth : %s" % (indent, datetime.fromtimestamp(self['authtime']).isoformat())))
+        print(("%sStart: %s" % (indent, datetime.fromtimestamp(self['starttime']).isoformat())))
+        print(("%sEnd  : %s" % (indent, datetime.fromtimestamp(self['endtime']).isoformat())))
+        print(("%sRenew: %s" % (indent, datetime.fromtimestamp(self['renew_till']).isoformat())))
 
 class Address(Structure):
     structure = (
@@ -116,7 +146,7 @@ class Principal:
         for i in self.components:
             totalLen += len(i)
         return totalLen
- 
+
     def getData(self):
         data = self.header.getData() + self.realm.getData()
         for component in self.components:
@@ -127,12 +157,20 @@ class Principal:
         return self.getData()
 
     def prettyPrint(self):
-        principal = ''
+        principal = b''
         for component in self.components:
-            principal += component['data'] + '/'
-        
+            if isinstance(component['data'], bytes) is not True:
+                component = b(component['data'])
+            else:
+                component = component['data']
+            principal += component + b'/'
+
         principal = principal[:-1]
-        principal += '@' + self.realm['data']
+        if isinstance(self.realm['data'], bytes):
+            realm = self.realm['data']
+        else:
+            realm = b(self.realm['data'])
+        principal += b'@' + realm
         return principal
 
     def fromPrincipal(self, principal):
@@ -153,18 +191,29 @@ class Principal:
         return types.Principal(self.prettyPrint(), type=self.header['name_type'])
 
 class Credential:
-    class CredentialHeader(Structure):
+    class CredentialHeaderV3(Structure):
         structure = (
             ('client',':', Principal),
             ('server',':', Principal),
-            ('key',':', KeyBlock),
+            ('key',':', KeyBlockV3),
             ('time',':', Times),
             ('is_skey','B=0'),
             ('tktflags','!L=0'),
             ('num_address','!L=0'),
         )
 
-    def __init__(self, data=None):
+    class CredentialHeaderV4(Structure):
+        structure = (
+            ('client',':', Principal),
+            ('server',':', Principal),
+            ('key',':', KeyBlockV4),
+            ('time',':', Times),
+            ('is_skey','B=0'),
+            ('tktflags','!L=0'),
+            ('num_address','!L=0'),
+        )
+
+    def __init__(self, data=None, ccache_version=None):
         self.addresses = ()
         self.authData = ()
         self.header = None
@@ -172,7 +221,11 @@ class Credential:
         self.secondTicket = None
 
         if data is not None:
-            self.header = self.CredentialHeader(data)
+            if ccache_version == 3:
+                self.header = self.CredentialHeaderV3(data)
+            else:
+                self.header = self.CredentialHeaderV4(data)
+
             data = data[len(self.header):]
             self.addresses = []
             for address in range(self.header['num_address']):
@@ -190,10 +243,10 @@ class Credential:
             self.secondTicket = CountedOctetString(data)
             data = data[len( self.secondTicket):]
         else:
-            self.header = self.CredentialHeader()
+            self.header = self.CredentialHeaderV4()
 
     def __getitem__(self, key):
-        return self.header[key] 
+        return self.header[key]
 
     def __setitem__(self, item, value):
         self.header[item] = value
@@ -202,7 +255,7 @@ class Credential:
         return self.header['server'].prettyPrint()
 
     def __len__(self):
-        totalLen = len(self.header) 
+        totalLen = len(self.header)
         for i in self.addresses:
             totalLen += len(i)
         totalLen += calcsize('!L')
@@ -211,7 +264,7 @@ class Credential:
         totalLen += len(self.ticket)
         totalLen += len(self.secondTicket)
         return totalLen
- 
+
     def dump(self):
         self.header.dump()
 
@@ -230,21 +283,21 @@ class Credential:
         return self.getData()
 
     def prettyPrint(self, indent=''):
-        print "%sClient: %s" % (indent, self.header['client'].prettyPrint())
-        print "%sServer: %s" % (indent, self.header['server'].prettyPrint())
-        print "%s%s" % (indent, self.header['key'].prettyPrint())
-        print "%sTimes: " % indent
+        print(("%sClient: %s" % (indent, self.header['client'].prettyPrint())))
+        print(("%sServer: %s" % (indent, self.header['server'].prettyPrint())))
+        print(("%s%s" % (indent, self.header['key'].prettyPrint())))
+        print(("%sTimes: " % indent))
         self.header['time'].prettyPrint('\t\t')
-        print "%sSubKey: %s" % (indent, self.header['is_skey'])
-        print "%sFlags: 0x%x" % (indent, self.header['tktflags'])
-        print "%sAddresses: %d" % (indent, self.header['num_address'])
+        print(("%sSubKey: %s" % (indent, self.header['is_skey'])))
+        print(("%sFlags: 0x%x" % (indent, self.header['tktflags'])))
+        print(("%sAddresses: %d" % (indent, self.header['num_address'])))
         for address in self.addresses:
             address.prettyPrint('\t\t')
-        print "%sAuth Data: %d" % (indent, len(self.authData))
+        print(("%sAuth Data: %d" % (indent, len(self.authData))))
         for ad in self.authData:
             ad.prettyPrint('\t\t')
-        print "%sTicket: %s" % (indent, self.ticket.prettyPrint())
-        print "%sSecond Ticket: %s" % (indent, self.secondTicket.prettyPrint())
+        print(("%sTicket: %s" % (indent, self.ticket.prettyPrint())))
+        print(("%sSecond Ticket: %s" % (indent, self.secondTicket.prettyPrint())))
 
     def toTGT(self):
         tgt_rep = AS_REP()
@@ -255,7 +308,7 @@ class Credential:
         # Fake EncryptedData
         tgt_rep['enc-part'] = noValue
         tgt_rep['enc-part']['etype'] = 1
-        tgt_rep['enc-part']['cipher'] = '' 
+        tgt_rep['enc-part']['cipher'] = ''
         seq_set(tgt_rep, 'cname', self['client'].toPrincipal().components_to_asn1)
         ticket = types.Ticket()
         ticket.from_asn1(self.ticket['data'])
@@ -266,9 +319,9 @@ class Credential:
         tgt = dict()
         tgt['KDC_REP'] = encoder.encode(tgt_rep)
         tgt['cipher'] = cipher
-        tgt['sessionKey'] = crypto.Key(cipher.enctype, str(self['key']['keyvalue']))
+        tgt['sessionKey'] = crypto.Key(cipher.enctype, self['key']['keyvalue'])
         return tgt
-        
+
     def toTGS(self, newSPN=None):
         tgs_rep = TGS_REP()
         tgs_rep['pvno'] = 5
@@ -278,7 +331,7 @@ class Credential:
         # Fake EncryptedData
         tgs_rep['enc-part'] = noValue
         tgs_rep['enc-part']['etype'] = 1
-        tgs_rep['enc-part']['cipher'] = '' 
+        tgs_rep['enc-part']['cipher'] = ''
         seq_set(tgs_rep, 'cname', self['client'].toPrincipal().components_to_asn1)
         ticket = types.Ticket()
         ticket.from_asn1(self.ticket['data'])
@@ -293,10 +346,13 @@ class Credential:
         tgs = dict()
         tgs['KDC_REP'] = encoder.encode(tgs_rep)
         tgs['cipher'] = cipher
-        tgs['sessionKey'] = crypto.Key(cipher.enctype, str(self['key']['keyvalue']))
+        tgs['sessionKey'] = crypto.Key(cipher.enctype, self['key']['keyvalue'])
         return tgs
-        
+
+
 class CCache:
+    # https://web.mit.edu/kerberos/krb5-devel/doc/formats/ccache_file_format.html
+
     class MiniHeader(Structure):
         structure = (
             ('file_format_version','!H=0x0504'),
@@ -308,29 +364,44 @@ class CCache:
         self.principal = None
         self.credentials = []
         self.miniHeader = None
+
         if data is not None:
-            miniHeader = self.MiniHeader(data)
-            data = data[len(str(miniHeader)):]
+            if PY2:
+                ccache_version = unpack('>B', data[1])[0]
+            else:
+                ccache_version = data[1]
 
-            headerLen = miniHeader['headerlen']
+            # Versions 1 and 2 are not implemented yet
+            if ccache_version == 1 or ccache_version == 2:
+                raise NotImplementedError('Not Implemented!')
 
-            self.headers = []
-            while headerLen > 0:
-                header = Header(data)
-                self.headers.append(header)
-                headerLen -= len(header)
-                data = data[len(header):]
+            # Only Version 4 contains a header
+            if ccache_version == 4:
+                miniHeader = self.MiniHeader(data)
+                data = data[len(miniHeader.getData()):]
+
+                headerLen = miniHeader['headerlen']
+
+                self.headers = []
+                while headerLen > 0:
+                    header = Header(data)
+                    self.headers.append(header)
+                    headerLen -= len(header)
+                    data = data[len(header):]
+            else:
+                # Skip over the version bytes
+                data = data[2:]
 
             # Now the primary_principal
             self.principal = Principal(data)
- 
+
             data = data[len(self.principal):]
-        
+
             # Now let's parse the credentials
             self.credentials = []
             while len(data) > 0:
-                cred = Credential(data)
-                if cred['server'].prettyPrint().find('krb5_ccache_conf_data') < 0:
+                cred = Credential(data, ccache_version)
+                if cred['server'].prettyPrint().find(b'krb5_ccache_conf_data') < 0:
                     self.credentials.append(cred)
                 data = data[len(cred.getData()):]
 
@@ -345,9 +416,9 @@ class CCache:
 
     def getCredential(self, server, anySPN=True):
         for c in self.credentials:
-            if c['server'].prettyPrint().upper() == server.upper() or c['server'].prettyPrint().upper().split('@')[0] == server.upper() \
-                    or c['server'].prettyPrint().upper().split('@')[0] == server.upper().split('@')[0]:
-                LOG.debug('Returning cached credential for %s' % c['server'].prettyPrint().upper())
+            if c['server'].prettyPrint().upper() == b(server.upper()) or c['server'].prettyPrint().upper().split(b'@')[0] == b(server.upper())\
+                    or c['server'].prettyPrint().upper().split(b'@')[0] == b(server.upper().split('@')[0]):
+                LOG.debug('Returning cached credential for %s' % c['server'].prettyPrint().upper().decode('utf-8'))
                 return c
         LOG.debug('SPN %s not found in cache' % server.upper())
         if anySPN is True:
@@ -355,14 +426,13 @@ class CCache:
             for c in self.credentials:
                 # Let's search for any TGT/TGS that matches the server w/o the SPN's service type/port, returns
                 # the first one
-                if c['server'].prettyPrint().find('/') >=0:
+                if c['server'].prettyPrint().find(b'/') >=0:
                     # Let's take the port out for comparison
-                    cachedSPN = '%s@%s'  % (c['server'].prettyPrint().upper().split('/')[1].split('@')[0].split(':')[0],
-                                               c['server'].prettyPrint().upper().split('/')[1].split('@')[1])
+                    cachedSPN = (c['server'].prettyPrint().upper().split(b'/')[1].split(b'@')[0].split(b':')[0] + b'@' + c['server'].prettyPrint().upper().split(b'/')[1].split(b'@')[1])
                     searchSPN = '%s@%s' % (server.upper().split('/')[1].split('@')[0].split(':')[0],
                                                server.upper().split('/')[1].split('@')[1])
-                    if cachedSPN == searchSPN:
-                        LOG.debug('Returning cached credential for %s' % c['server'].prettyPrint().upper())
+                    if cachedSPN == b(searchSPN):
+                        LOG.debug('Returning cached credential for %s' % c['server'].prettyPrint().upper().decode('utf-8'))
                         return c
 
         return None
@@ -370,7 +440,7 @@ class CCache:
     def toTimeStamp(self, dt, epoch=datetime(1970,1,1)):
         td = dt - epoch
         # return td.total_seconds()
-        return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 1e6
+        return int((td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) // 1e6)
 
     def reverseFlags(self, flags):
         result = 0
@@ -386,7 +456,7 @@ class CCache:
         header = Header()
         header['tag'] = 1
         header['taglen'] = 8
-        header['tagdata'] = '\xff\xff\xff\xff\x00\x00\x00\x00'
+        header['tagdata'] = b'\xff\xff\xff\xff\x00\x00\x00\x00'
         self.headers.append(header)
 
         decodedTGT = decoder.decode(tgt, asn1Spec = AS_REP())[0]
@@ -405,31 +475,31 @@ class CCache:
         # AS-REP encrypted part (includes TGS session key or
         # application session key), encrypted with the client key
         # (Section 5.4.2)
-        plainText = cipher.decrypt(oldSessionKey, 3, str(cipherText))
+        plainText = cipher.decrypt(oldSessionKey, 3, cipherText)
 
         encASRepPart = decoder.decode(plainText, asn1Spec = EncASRepPart())[0]
-
         credential = Credential()
         server = types.Principal()
         server.from_asn1(encASRepPart, 'srealm', 'sname')
         tmpServer = Principal()
         tmpServer.fromPrincipal(server)
-        
+
         credential['client'] = self.principal
         credential['server'] = tmpServer
         credential['is_skey'] = 0
 
-        credential['key'] = KeyBlock()
+        credential['key'] = KeyBlockV4()
         credential['key']['keytype'] = int(encASRepPart['key']['keytype'])
-        credential['key']['keyvalue'] = str(encASRepPart['key']['keyvalue'])
+        credential['key']['keyvalue'] = encASRepPart['key']['keyvalue'].asOctets()
         credential['key']['keylen'] = len(credential['key']['keyvalue'])
 
         credential['time'] = Times()
         credential['time']['authtime'] = self.toTimeStamp(types.KerberosTime.from_asn1(encASRepPart['authtime']))
-        credential['time']['starttime'] = self.toTimeStamp(types.KerberosTime.from_asn1(encASRepPart['starttime'])) 
+        credential['time']['starttime'] = self.toTimeStamp(types.KerberosTime.from_asn1(encASRepPart['starttime']))
         credential['time']['endtime'] = self.toTimeStamp(types.KerberosTime.from_asn1(encASRepPart['endtime']))
-        credential['time']['renew_till'] = self.toTimeStamp(types.KerberosTime.from_asn1(encASRepPart['renew-till'])) 
-
+        # After KB4586793 for CVE-2020-17049 this timestamp may be omitted
+        if encASRepPart['renew-till'].hasValue():
+            credential['time']['renew_till'] = self.toTimeStamp(types.KerberosTime.from_asn1(encASRepPart['renew-till']))
         flags = self.reverseFlags(encASRepPart['flags'])
         credential['tktflags'] = flags
 
@@ -438,7 +508,7 @@ class CCache:
         credential.ticket['data'] = encoder.encode(decodedTGT['ticket'].clone(tagSet=Ticket.tagSet, cloneValueFlag=True))
         credential.ticket['length'] = len(credential.ticket['data'])
         credential.secondTicket = CountedOctetString()
-        credential.secondTicket['data'] = ''
+        credential.secondTicket['data'] = b''
         credential.secondTicket['length'] = 0
         self.credentials.append(credential)
 
@@ -447,7 +517,7 @@ class CCache:
         header = Header()
         header['tag'] = 1
         header['taglen'] = 8
-        header['tagdata'] = '\xff\xff\xff\xff\x00\x00\x00\x00'
+        header['tagdata'] = b'\xff\xff\xff\xff\x00\x00\x00\x00'
         self.headers.append(header)
 
         decodedTGS = decoder.decode(tgs, asn1Spec = TGS_REP())[0]
@@ -465,7 +535,7 @@ class CCache:
         # Key Usage 8
         # TGS-REP encrypted part (includes application session
         # key), encrypted with the TGS session key (Section 5.4.2)
-        plainText = cipher.decrypt(oldSessionKey, 8, str(cipherText))
+        plainText = cipher.decrypt(oldSessionKey, 8, cipherText)
 
         encTGSRepPart = decoder.decode(plainText, asn1Spec = EncTGSRepPart())[0]
 
@@ -474,21 +544,23 @@ class CCache:
         server.from_asn1(encTGSRepPart, 'srealm', 'sname')
         tmpServer = Principal()
         tmpServer.fromPrincipal(server)
-        
+
         credential['client'] = self.principal
         credential['server'] = tmpServer
-        credential['is_skey'] = 0 
+        credential['is_skey'] = 0
 
-        credential['key'] = KeyBlock()
+        credential['key'] = KeyBlockV4()
         credential['key']['keytype'] = int(encTGSRepPart['key']['keytype'])
-        credential['key']['keyvalue'] = str(encTGSRepPart['key']['keyvalue'])
+        credential['key']['keyvalue'] = encTGSRepPart['key']['keyvalue'].asOctets()
         credential['key']['keylen'] = len(credential['key']['keyvalue'])
 
         credential['time'] = Times()
         credential['time']['authtime'] = self.toTimeStamp(types.KerberosTime.from_asn1(encTGSRepPart['authtime']))
-        credential['time']['starttime'] = self.toTimeStamp(types.KerberosTime.from_asn1(encTGSRepPart['starttime'])) 
+        credential['time']['starttime'] = self.toTimeStamp(types.KerberosTime.from_asn1(encTGSRepPart['starttime']))
         credential['time']['endtime'] = self.toTimeStamp(types.KerberosTime.from_asn1(encTGSRepPart['endtime']))
-        credential['time']['renew_till'] = self.toTimeStamp(types.KerberosTime.from_asn1(encTGSRepPart['renew-till'])) 
+        # After KB4586793 for CVE-2020-17049 this timestamp may be omitted
+        if encTGSRepPart['renew-till'].hasValue():
+            credential['time']['renew_till'] = self.toTimeStamp(types.KerberosTime.from_asn1(encTGSRepPart['renew-till']))
 
         flags = self.reverseFlags(encTGSRepPart['flags'])
         credential['tktflags'] = flags
@@ -499,31 +571,211 @@ class CCache:
         credential.ticket['data'] = encoder.encode(decodedTGS['ticket'].clone(tagSet=Ticket.tagSet, cloneValueFlag=True))
         credential.ticket['length'] = len(credential.ticket['data'])
         credential.secondTicket = CountedOctetString()
-        credential.secondTicket['data'] = ''
+        credential.secondTicket['data'] = b''
         credential.secondTicket['length'] = 0
         self.credentials.append(credential)
 
     @classmethod
     def loadFile(cls, fileName):
-        f = open(fileName,'rb')
-        data = f.read()
-        f.close()
-        return cls(data)
+        if fileName is None:
+            LOG.critical('CCache file is not found. Skipping...')
+            LOG.debug('The specified path is not correct or the KRB5CCNAME environment variable is not defined')
+            return None
+
+        try:
+            f = open(fileName, 'rb')
+            data = f.read()
+            f.close()
+            return cls(data)
+        except FileNotFoundError as e:
+            raise e
 
     def saveFile(self, fileName):
-        f = open(fileName,'wb+')
+        f = open(fileName, 'wb+')
         f.write(self.getData())
         f.close()
 
+    @classmethod
+    def parseFile(cls, domain='', username='', target=''):
+        """
+        parses the CCache file specified in the KRB5CCNAME environment variable
+
+        :param domain: an optional domain name of a user
+        :param username: an optional username of a user
+        :param target: an optional SPN of a target system
+
+        :return: domain, username, TGT, TGS
+        """
+
+        ccache = cls.loadFile(os.getenv('KRB5CCNAME'))
+        if ccache is None:
+            return domain, username, None, None
+
+        LOG.debug('Using Kerberos Cache: %s' % os.getenv('KRB5CCNAME'))
+
+        if domain == '':
+            domain = ccache.principal.realm['data'].decode('utf-8')
+            LOG.debug('Domain retrieved from CCache: %s' % domain)
+
+        creds = None
+        if target != '':
+            principal = '%s@%s' % (target.upper(), domain.upper())
+            creds = ccache.getCredential(principal)
+
+        TGT = None
+        TGS = None
+        if creds is None:
+            principal = 'krbtgt/%s@%s' % (domain.upper(), domain.upper())
+            creds = ccache.getCredential(principal)
+            if creds is not None:
+                LOG.debug('Using TGT from cache')
+                TGT = creds.toTGT()
+            else:
+                LOG.debug('No valid credentials found in cache')
+        else:
+            LOG.debug('Using TGS from cache')
+            TGS = creds.toTGS(principal)
+
+        if username == '' and creds is not None:
+            username = creds['client'].prettyPrint().split(b'@')[0].decode('utf-8')
+            LOG.debug('Username retrieved from CCache: %s' % username)
+        elif username == '' and len(ccache.principal.components) > 0:
+            username = ccache.principal.components[0]['data'].decode('utf-8')
+            LOG.debug('Username retrieved from CCache: %s' % username)
+
+        return domain, username, TGT, TGS
+
     def prettyPrint(self):
-        print "Primary Principal: %s" % self.principal.prettyPrint()
-        print "Credentials: "
+        print(("Primary Principal: %s" % self.principal.prettyPrint()))
+        print("Credentials: ")
         for i, credential in enumerate(self.credentials):
-            print "[%d]" % i
-            credential.prettyPrint('\t') 
+            print(("[%d]" % i))
+            credential.prettyPrint('\t')
+
+    @classmethod
+    def loadKirbiFile(cls, fileName):
+        f = open(fileName, 'rb')
+        data = f.read()
+        f.close()
+        ccache = cls()
+        ccache.fromKRBCRED(data)
+        return ccache
+
+    def saveKirbiFile(self, fileName):
+        f = open(fileName, 'wb+')
+        f.write(self.toKRBCRED())
+        f.close()
+
+    def fromKRBCRED(self, encodedKrbCred):
+
+        krbCred = decoder.decode(encodedKrbCred, asn1Spec=KRB_CRED())[0]
+        encKrbCredPart = decoder.decode(krbCred['enc-part']['cipher'], asn1Spec=EncKrbCredPart())[0]
+        krbCredInfo = encKrbCredPart['ticket-info'][0]
+
+        self.setDefaultHeader()
+
+        tmpPrincipal = types.Principal()
+        tmpPrincipal.from_asn1(krbCredInfo, 'prealm', 'pname')
+        self.principal = Principal()
+        self.principal.fromPrincipal(tmpPrincipal)
+
+        credential = Credential()
+        server = types.Principal()
+        server.from_asn1(krbCredInfo, 'srealm', 'sname')
+        tmpServer = Principal()
+        tmpServer.fromPrincipal(server)
+
+        credential['client'] = self.principal
+        credential['server'] = tmpServer
+        credential['is_skey'] = 0
+
+        credential['key'] = KeyBlockV4()
+        credential['key']['keytype'] = int(krbCredInfo['key']['keytype'])
+        credential['key']['keyvalue'] = krbCredInfo['key']['keyvalue'].asOctets()
+        credential['key']['keylen'] = len(credential['key']['keyvalue'])
+
+        credential['time'] = Times()
+
+        credential['time']['authtime'] = self.toTimeStamp(types.KerberosTime.from_asn1(krbCredInfo['starttime']))
+        credential['time']['starttime'] = self.toTimeStamp(types.KerberosTime.from_asn1(krbCredInfo['starttime']))
+        credential['time']['endtime'] = self.toTimeStamp(types.KerberosTime.from_asn1(krbCredInfo['endtime']))
+        # After KB4586793 for CVE-2020-17049 this timestamp may be omitted
+        if krbCredInfo['renew-till'].hasValue():
+            credential['time']['renew_till'] = self.toTimeStamp(types.KerberosTime.from_asn1(krbCredInfo['renew-till']))
+
+        flags = self.reverseFlags(krbCredInfo['flags'])
+        credential['tktflags'] = flags
+
+        credential['num_address'] = 0
+        credential.ticket = CountedOctetString()
+        credential.ticket['data'] = encoder.encode(
+            krbCred['tickets'][0].clone(tagSet=Ticket.tagSet, cloneValueFlag=True)
+        )
+        credential.ticket['length'] = len(credential.ticket['data'])
+        credential.secondTicket = CountedOctetString()
+        credential.secondTicket['data'] = b''
+        credential.secondTicket['length'] = 0
+
+        self.credentials.append(credential)
+
+    def toKRBCRED(self):
+        principal = self.principal
+        credential = self.credentials[0]
+
+        krbCredInfo = KrbCredInfo()
+
+        krbCredInfo['key'] = noValue
+        krbCredInfo['key']['keytype'] = credential['key']['keytype']
+        krbCredInfo['key']['keyvalue'] = credential['key']['keyvalue']
+
+        krbCredInfo['prealm'] = principal.realm.fields['data']
+
+        krbCredInfo['pname'] = noValue
+        krbCredInfo['pname']['name-type'] = principal.header['name_type']
+        seq_set_iter(krbCredInfo['pname'], 'name-string', (principal.components[0].fields['data'],))
+
+        krbCredInfo['flags'] = credential['tktflags']
+
+        krbCredInfo['starttime'] = KerberosTime.to_asn1(datetime.utcfromtimestamp(credential['time']['starttime']))
+        krbCredInfo['endtime'] = KerberosTime.to_asn1(datetime.utcfromtimestamp(credential['time']['endtime']))
+        krbCredInfo['renew-till'] = KerberosTime.to_asn1(datetime.utcfromtimestamp(credential['time']['renew_till']))
+
+        krbCredInfo['srealm'] = credential['server'].realm.fields['data']
+
+        krbCredInfo['sname'] = noValue
+        krbCredInfo['sname']['name-type'] = credential['server'].header['name_type']
+        tmp_service_class = credential['server'].components[0].fields['data']
+        tmp_service_hostname = credential['server'].components[1].fields['data']
+        seq_set_iter(krbCredInfo['sname'], 'name-string', (tmp_service_class, tmp_service_hostname))
+
+        encKrbCredPart = EncKrbCredPart()
+        seq_set_iter(encKrbCredPart, 'ticket-info', (krbCredInfo,))
+
+        krbCred = KRB_CRED()
+        krbCred['pvno'] = 5
+        krbCred['msg-type'] = 22
+
+        krbCred['enc-part'] = noValue
+        krbCred['enc-part']['etype'] = 0
+        krbCred['enc-part']['cipher'] = encoder.encode(encKrbCredPart)
+
+        ticket = decoder.decode(credential.ticket['data'], asn1Spec=Ticket())[0]
+        seq_set_iter(krbCred, 'tickets', (ticket,))
+
+        encodedKrbCred = encoder.encode(krbCred)
+
+        return encodedKrbCred
+
+    def setDefaultHeader(self):
+        self.headers = []
+        header = Header()
+        header['tag'] = 1
+        header['taglen'] = 8
+        header['tagdata'] = b'\xff\xff\xff\xff\x00\x00\x00\x00'
+        self.headers.append(header)
+
 
 
 if __name__ == '__main__':
-    import os
     ccache = CCache.loadFile(os.getenv('KRB5CCNAME'))
     ccache.prettyPrint()
