@@ -1,14 +1,22 @@
-# SECUREAUTH LABS. Copyright 2018 SecureAuth Corporation. All rights reserved.
+# Impacket - Collection of Python classes for working with network protocols.
 #
-# This software is provided under under a slightly modified version
+# Copyright Fortra, LLC and its affiliated companies 
+#
+# All rights reserved.
+#
+# This software is provided under a slightly modified version
 # of the Apache Software License. See the accompanying LICENSE file
 # for more information.
 #
-# Author: Alberto Solino (@agsolino)
-#
 # Description:
 #   SMB 2 and 3 Protocol Structures and constants [MS-SMB2]
 #
+# Author:
+#   Alberto Solino (@agsolino)
+#
+
+from __future__ import division
+from __future__ import print_function
 
 from impacket.structure import Structure
 
@@ -55,6 +63,23 @@ SYMLINK_FLAG_RELATIVE         = 0x1
 SMB2_NEGOTIATE_SIGNING_ENABLED  = 0x1
 SMB2_NEGOTIATE_SIGNING_REQUIRED = 0x2
 
+# SMB2_NEGOTIATE_CONTEXT
+SMB2_PREAUTH_INTEGRITY_CAPABILITIES = 0x1
+SMB2_ENCRYPTION_CAPABILITIES        = 0x2
+SMB2_COMPRESSION_CAPABILITIES       = 0x3
+SMB2_NETNAME_NEGOTIATE_CONTEXT_ID   = 0x5
+
+# SMB2_COMPRESSION_CAPABILITIES
+SMB2_COMPRESSION_CAPABILITIES_FLAG_NONE    = 0x0
+SMB2_COMPRESSION_CAPABILITIES_FLAG_CHAINED = 0x1
+
+# Compression Algorithms
+COMPRESSION_ALGORITHM_NONE         = 0x0
+COMPRESSION_ALGORITHM_LZNT1        = 0x1
+COMPRESSION_ALGORITHM_LZ77         = 0x2
+COMPRESSION_ALGORITHM_LZ77_HUFFMAN = 0x3
+COMPRESSION_ALGORITHM_PATTERN_V1   = 0x4
+
 # Capabilities
 SMB2_GLOBAL_CAP_DFS                = 0x01
 SMB2_GLOBAL_CAP_LEASING            = 0x02
@@ -65,12 +90,12 @@ SMB2_GLOBAL_CAP_DIRECTORY_LEASING  = 0x20
 SMB2_GLOBAL_CAP_ENCRYPTION         = 0x40
 
 # Dialects
-SMB2_DIALECT_002      = 0x0202 
-SMB2_DIALECT_21       = 0x0210 
-SMB2_DIALECT_30       = 0x0300 
+SMB2_DIALECT_002      = 0x0202
+SMB2_DIALECT_21       = 0x0210
+SMB2_DIALECT_30       = 0x0300
 SMB2_DIALECT_302      = 0x0302  #SMB 3.0.2
 SMB2_DIALECT_311      = 0x0311  #SMB 3.1.1
-SMB2_DIALECT_WILDCARD = 0x02FF 
+SMB2_DIALECT_WILDCARD = 0x02FF
 
 # SMB2_SESSION_SETUP
 # Flags
@@ -144,7 +169,7 @@ FILE_SHARE_WRITE        = 0x00000002
 FILE_SHARE_DELETE       = 0x00000004
 
 # Create Disposition
-FILE_SUPERSEDE          = 0x00000000 
+FILE_SUPERSEDE          = 0x00000000
 FILE_OPEN               = 0x00000001
 FILE_CREATE             = 0x00000002
 FILE_OPEN_IF            = 0x00000003
@@ -167,7 +192,7 @@ FILE_OPEN_BY_FILE_ID           = 0x00002000
 FILE_OPEN_FOR_BACKUP_INTENT    = 0x00004000
 FILE_NO_COMPRESSION            = 0x00008000
 FILE_RESERVE_OPFILTER          = 0x00100000
-FILE_OPEN_REPARSE_POINT        = 0x00200000 
+FILE_OPEN_REPARSE_POINT        = 0x00200000
 FILE_OPEN_NO_RECALL            = 0x00400000
 FILE_OPEN_FOR_FREE_SPACE_QUERY = 0x00800000
 
@@ -200,19 +225,19 @@ FILE_TRAVERSE          = 0x00000020
 FILE_DELETE_CHILD      = 0x00000040
 
 # Create Contexts
-SMB2_CREATE_EA_BUFFER                     = 0x45787441 
+SMB2_CREATE_EA_BUFFER                     = 0x45787441
 SMB2_CREATE_SD_BUFFER                     = 0x53656344
-SMB2_CREATE_DURABLE_HANDLE_REQUEST        = 0x44486e51 
-SMB2_CREATE_DURABLE_HANDLE_RECONNECT      = 0x44486e43 
-SMB2_CREATE_ALLOCATION_SIZE               = 0x416c5369 
-SMB2_CREATE_QUERY_MAXIMAL_ACCESS_REQUEST  = 0x4d784163 
-SMB2_CREATE_TIMEWARP_TOKEN                = 0x54577270 
-SMB2_CREATE_QUERY_ON_DISK_ID              = 0x51466964 
-SMB2_CREATE_REQUEST                       = 0x52714c73 
-SMB2_CREATE_REQUEST_LEASE_V2              = 0x52714c73 
-SMB2_CREATE_DURABLE_HANDLE_REQUEST_V2     = 0x44483251 
-SMB2_CREATE_DURABLE_HANDLE_RECONNECT_V2   = 0x44483243 
-SMB2_CREATE_APP_INSTANCE_ID               = 0x45BCA66AEFA7F74A9008FA462E144D74 
+SMB2_CREATE_DURABLE_HANDLE_REQUEST        = 0x44486e51
+SMB2_CREATE_DURABLE_HANDLE_RECONNECT      = 0x44486e43
+SMB2_CREATE_ALLOCATION_SIZE               = 0x416c5369
+SMB2_CREATE_QUERY_MAXIMAL_ACCESS_REQUEST  = 0x4d784163
+SMB2_CREATE_TIMEWARP_TOKEN                = 0x54577270
+SMB2_CREATE_QUERY_ON_DISK_ID              = 0x51466964
+SMB2_CREATE_REQUEST                       = 0x52714c73
+SMB2_CREATE_REQUEST_LEASE_V2              = 0x52714c73
+SMB2_CREATE_DURABLE_HANDLE_REQUEST_V2     = 0x44483251
+SMB2_CREATE_DURABLE_HANDLE_RECONNECT_V2   = 0x44483243
+SMB2_CREATE_APP_INSTANCE_ID               = 0x45BCA66AEFA7F74A9008FA462E144D74
 
 # Flags
 SMB2_CREATE_FLAG_REPARSEPOINT  = 0x1
@@ -235,7 +260,7 @@ SMB2_LEASE_FLAG_PARENT_LEASE_KEY_SET = 0x4
 
 # SMB2_CREATE_DURABLE_HANDLE_REQUEST_V2 Flags
 SMB2_DHANDLE_FLAG_PERSISTENT = 0x02
- 
+
 # SMB2_CLOSE
 # Flags
 SMB2_CLOSE_FLAG_POSTQUERY_ATTRIB  = 0x0001
@@ -299,7 +324,7 @@ RSS_CAPABLE  = 0x01
 RDMA_CAPABLE = 0x02
 
 # SMB2_QUERY_DIRECTORIES
-# Information Class 
+# Information Class
 FILE_DIRECTORY_INFORMATION         = 0x01
 FILE_FULL_DIRECTORY_INFORMATION    = 0x02
 FILEID_FULL_DIRECTORY_INFORMATION  = 0x26
@@ -336,7 +361,7 @@ FILE_NOTIFY_CHANGE_STREAM_WRITE  = 0x00000800
 FILE_ACTION_ADDED            = 0x00000001
 FILE_ACTION_REMOVED          = 0x00000002
 FILE_ACTION_MODIFIED         = 0x00000003
-FILE_ACTION_RENAMED_OLD_NAME = 0x00000004 
+FILE_ACTION_RENAMED_OLD_NAME = 0x00000004
 FILE_ACTION_RENAMED_NEW_NAME = 0x00000005
 
 # SMB2_QUERY_INFO
@@ -431,7 +456,7 @@ class SMBPacketBase(Structure):
 
     def isValidAnswer(self, status):
         if self['Status'] != status:
-            import smb3
+            from . import smb3
             raise smb3.SessionError(self['Status'], self)
         return True
 
@@ -546,16 +571,26 @@ class SMB2Negotiate(Structure):
         ('Reserved','<H=0'),
         ('Capabilities','<L=0'),
         ('ClientGuid','16s=""'),
-        ('ClientStartTime','<Q=0'),
+        ('ClientStartTime','8s=""'),  # or (NegotiateContextOffset/NegotiateContextCount/Reserved2) in SMB 3.1.1
         ('Dialects','*<H'),
+        # SMB 3.1.1
+        ('Padding',':=""'),
+        ('NegotiateContextList',':=""'),
     )
 
+class SMB311ContextData(Structure):
+    structure = (
+        ('NegotiateContextOffset','<L=0'),
+        ('NegotiateContextCount','<H=0'),
+        ('Reserved2','<H=0'),
+    )
 class SMB2Negotiate_Response(Structure):
     structure = (
         ('StructureSize','<H=65'),
         ('SecurityMode','<H=0'),
         ('DialectRevision','<H=0'),
-        ('Reserved','<H=0'),
+        # SMB 3.1.1 only. Otherwise Reserved
+        ('NegotiateContextCount','<H=0'),
         ('ServerGuid','16s=""'),
         ('Capabilities','<L=0'),
         ('MaxTransactSize','<L=0'),
@@ -565,14 +600,63 @@ class SMB2Negotiate_Response(Structure):
         ('ServerStartTime','<Q=0'),
         ('SecurityBufferOffset','<H=0'),
         ('SecurityBufferLength','<H=0'),
-        ('Reserved2','<L=0'),
+        # SMB 3.1.1 only. Otherwise Reserved
+        ('NegotiateContextOffset','<L=0'),
         ('_AlignPad','_-AlignPad','self["SecurityBufferOffset"] - (64 + self["StructureSize"] - 1)'),
         ('AlignPad',':=""'),
         ('_Buffer','_-Buffer','self["SecurityBufferLength"]'),
         ('Buffer',':'),
+        ('_Padding','_-Padding', '0 if self["NegotiateContextOffset"] == 0 else (self["NegotiateContextOffset"] - '
+                                 'self["SecurityBufferOffset"] - self["SecurityBufferLength"])'),
+        ('Padding',':=""'),
+        ('_NegotiateContextList','_-NegotiateContextList', '0 if self["NegotiateContextOffset"] == 0 else '
+                                                           'len(self.rawData)-self["NegotiateContextOffset"]+64'),
+        ('NegotiateContextList',':=""'),
     )
 
-# SMB2_SESSION_SETUP 
+# SMB2 NEGOTIATE_CONTEXT
+class SMB2NegotiateContext(Structure):
+    structure = (
+        ('ContextType','<H=0'),
+        ('DataLength','<H=0'),
+        ('Reserved','<L=0'),
+        ('_Data', '_-Data', 'self["DataLength"]'),
+        ('Data',':=""'),
+    )
+
+# SMB2_PREAUTH_INTEGRITY_CAPABILITIES
+class SMB2PreAuthIntegrityCapabilities(Structure):
+    structure = (
+        ('HashAlgorithmCount','<H=0'),
+        ('SaltLength','<H=0'),
+        ('_HashAlgorithms', '_-HashAlgorithms', 'self["HashAlgorithmCount"]*2'),
+        ('HashAlgorithms',':=""'),
+        ('Salt',':=""'),
+    )
+
+# SMB2_ENCRYPTION_CAPABILITIES
+class SMB2EncryptionCapabilities(Structure):
+    structure = (
+        ('CipherCount','<H=0'),
+        ('Ciphers',':=""'),
+    )
+
+# SMB2_COMPRESSION_CAPABILITIES
+class SMB2CompressionCapabilities(Structure):
+    structure = (
+        ('CompressionAlgorithmCount','<H=0'),
+        ('Padding','<H=0'),
+        ('Flags','<L=0'),
+        ('CompressionAlgorithms',':=""'),
+    )
+
+# SMB2_NETNAME_NEGOTIATE_CONTEXT_ID
+class SMB2NetNameNegotiateContextID(Structure):
+    structure = (
+        ('NetName',':=""'),
+    )
+
+# SMB2_SESSION_SETUP
 class SMB2SessionSetup(Structure):
     SIZE = 24
     structure = (
@@ -597,10 +681,10 @@ class SMB2SessionSetup(Structure):
 
     def getData(self):
         #self['AlignPad'] = '\x00' * ((8 - ((24 + SMB2_PACKET_SIZE) & 7)) & 7)
-        #self['SecurityBufferOffset'] = 24 + SMB2_PACKET_SIZE +len(self['AlignPad']) 
+        #self['SecurityBufferOffset'] = 24 + SMB2_PACKET_SIZE +len(self['AlignPad'])
         #self['SecurityBufferLength'] += len(self['AlignPad'])
         return Structure.getData(self)
-        
+
 
 class SMB2SessionSetup_Response(Structure):
     structure = (
@@ -619,7 +703,7 @@ class SMB2Logoff(Structure):
     structure = (
         ('StructureSize','<H=4'),
         ('Reserved','<H=0'),
-    ) 
+    )
 
 
 class SMB2Logoff_Response(Structure):
@@ -785,7 +869,7 @@ class SMB2_CREATE_ALLOCATION_SIZE(Structure):
 
 class SMB2_CREATE_TIMEWARP_TOKEN(Structure):
     structure = (
-        ('AllocationSize','<Q=0'),
+        ('Timestamp','<Q=0'),
     )
 
 class SMB2_CREATE_REQUEST_LEASE(Structure):
@@ -845,7 +929,7 @@ class SMB2_CREATE_QUERY_ON_DISK_ID(Structure):
     )
 
 # Todo: Add Classes for
-#SMB2_CREATE_SD_BUFFER                    
+#SMB2_CREATE_SD_BUFFER
 
 # SMB2_CLOSE
 class SMB2Close(Structure):
@@ -903,7 +987,7 @@ class SMB2Read(Structure):
         ('_AlignPad','_-AlignPad','self["ReadChannelInfoOffset"] - (64 + self["StructureSize"] - 1)'),
         ('AlignPad',':=""'),
         ('_Buffer','_-Buffer','self["ReadChannelInfoLength"]'),
-        ('Buffer',':=0'),
+        ('Buffer',':="0"'),
     )
     def __init__(self, data = None):
         Structure.__init__(self,data)
@@ -1112,7 +1196,7 @@ class NETWORK_RESILIENCY_REQUEST(Structure):
     structure = (
         ('Timeout','<L=0'),
         ('Reserved','<L=0'),
-    ) 
+    )
 
 class VALIDATE_NEGOTIATE_INFO(Structure):
     structure = (
@@ -1342,6 +1426,76 @@ class SMB2QueryInfo_Response(Structure):
        ('Buffer',':'),
    )
 
+class FILE_BASIC_INFORMATION (Structure):
+    structure = (
+        ('CreationTime','<q'),
+        ('LastAccessTime','<q'),
+        ('LastWriteTime','<q'),
+        ('ChangeTime','<q'),
+        ('FileAttributes','<L'),
+        ('Reserved','<L=0'),
+    )
+
+class FILE_STANDARD_INFORMATION (Structure):
+    structure = (
+        ('AllocationSize','<q'),
+        ('EndOfFile','<q'),
+        ('NumberOfLinks','<L'),
+        ('DeletePending','<B=0'),
+        ('Directory','<B'),
+        ('Reserved','<H=0'),
+    )
+
+class FILE_INTERNAL_INFORMATION (Structure):
+    structure = (
+        ('IndexNumber','<q=0'),
+    )
+
+class FILE_EA_INFORMATION (Structure):
+    structure = (
+        ('EaSize','<L'),
+    )
+
+class FILE_ACCESS_INFORMATION (Structure):
+    structure = (
+        ('AccessFlags','<L'),
+    )
+
+class FILE_POSITION_INFORMATION (Structure):
+    structure = (
+        ('CurrentByteOffset','<Q'),
+    )
+
+class FILE_MODE_INFORMATION (Structure):
+    structure = (
+        ('Mode','<L=0'),
+    )
+
+class FILE_ALIGNMENT_INFORMATION (Structure):
+    structure = (
+        ('AlignmentRequirement','<L'),
+    )
+
+class FILE_NAME_INFORMATION (Structure):
+    structure = (
+        ('FileNameLength','<L=0'),
+        ('_FileName','_-FileName', 'self["FileNameLength"]'),
+        ('FileName',':'),
+    )
+
+class FILE_ALL_INFORMATION(Structure):
+    structure = (
+        ('BasicInformation',':',FILE_BASIC_INFORMATION),
+        ('StandardInformation',':',FILE_STANDARD_INFORMATION),
+        ('InternalInformation',':',FILE_INTERNAL_INFORMATION),
+        ('EaInformation',':',FILE_EA_INFORMATION),
+        ('AccessInformation',':',FILE_ACCESS_INFORMATION),
+        ('PositionInformation',':',FILE_POSITION_INFORMATION),
+        ('ModeInformation',':',FILE_MODE_INFORMATION),
+        ('AlignmentInformation',':',FILE_ALIGNMENT_INFORMATION),
+        ('NameInformation',':',FILE_NAME_INFORMATION),
+    )
+
 # SMB2_SET_INFO
 class SMB2SetInfo(Structure):
     SIZE = 32
@@ -1390,13 +1544,31 @@ class SMB2_TRANSFORM_HEADER(Structure):
         ('SessionID','<Q=0'),
     )
 
-# SMB2_FILE_INTERNAL_INFO
-class FileInternalInformation(Structure):
+class SMB2_COMPRESSION_TRANSFORM_HEADER(Structure):
     structure = (
-        ('IndexNumber','<q=0'),
+        ('ProtocolID','<L=0'),
+        ('OriginalCompressedSegmentSize','<L=0'),
+        ('CompressionAlgorithm','<H=0'),
+        ('Flags','<H=0'),
+        ('Offset_Length','<L=0'),
     )
 
-# SMB2_SEC_INFO_00       
+class SMB2_COMPRESSION_PAYLOAD_HEADER(Structure):
+    structure = (
+        ('AlgorithmId','<H=0'),
+        ('Reserved','<H=0'),
+        ('Length','<L=0'),
+    )
+
+class SMB2_COMPRESSION_PATTERN_PAYLOAD_V1(Structure):
+    structure = (
+        ('Pattern','B=0'),
+        ('Reserved1','B=0'),
+        ('Reserved2','B=0'),
+        ('Repetitions','<L=0'),
+    )
+
+# SMB2_SEC_INFO_00
 class FileSecInformation(Structure):
     structure = (
         ('Revision','<h=1'),
